"""NGPVAN Activist Code Endpoints"""

import logging
logger = logging.getLogger(__name__)


class ActivistCodes(object):
    """
        .. note::
            To apply or remove activist codes, use the :meth:`toggle_activist_code` method.
    """

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_activist_codes(self):
        """
        Get activist codes.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'activistCodes'

        acs = self.connection.request_paginate(url)
        logger.debug(acs)
        logger.info(f'Found {acs.num_rows} activist codes.')

        return acs

    def get_activist_code(self, activist_code_id):
        """Get an activist code.

        `Args:`
            activist_code_id : int
                The activist code id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'activistCodes/{}'.format(activist_code_id)

        ac = self.connection.request(url)
        logger.debug(ac)
        logger.info(f'Found activist code {activist_code_id}.')
        return ac



from tempfile import NamedTemporaryFile
import gzip

__all__ = [
    'create_temp_file',
    'create_temp_file_for_path',
    'is_gzip_path',
    'suffix_for_compression_type',
    'compression_type_for_path',
    'string_to_temp_file'
    ]


# This global list keeps track of all temp files created during the runtime of a script.
# We can't rely exclusively on the "automatic removal" behavior of the built-in `tempfile`
# library, because of our use of petl. Specifically, if a petl table is loaded from a
# temporary file (eg. a CSV), petl may not actually read the file until much later, after the
# TemporaryFile object has already gone out of scope and the file removed. If this
# occurs, the petl load will fail since it's trying to read from a file that doesn't exist.
# So by storing all the temp files in this list, we can keep them "in scope" for the remaining
# runtime of the script.
_temp_files = []


def create_temp_file(suffix=None):
    """
    Create a temp file that will exist as long as the current script is running.

    `Args:`
        suffix: str
            A suffix/extension to add to the end of the temp file name
    `Returns:`
        str
            The path of the temp file
    """
    temp_file = NamedTemporaryFile(suffix=suffix)
    _temp_files.append(temp_file)
    return temp_file.name


def create_temp_file_for_path(path):
    """
    Creates a temp file that will exist as long as the current script is running, and with
    a file name mimicking that of the provided path.

    `Args:`
        path: str
            Path (or just file name) of the file you want the temp file to mimick.
    `Returns:`
        str
            The path of the temp file
    """

    # Add the appropriate compression suffix to the file, so other libraries that check the
    # file's extension will know that it is compressed.
    # TODO Make this more robust, maybe even using the entire remote file name as the suffix.
    suffix = '.gz' if is_gzip_path(path) else None
    return create_temp_file(suffix=suffix)


def close_temp_file(path):
    """
    Force closes a Parsons temp file, which will cause it to be deleted immediately.

    Useful for when you don't want to wait until the end of your script's execution for temp
    files to be closed and deleted. Eg. If you're running into system limits on open file
    descriptors.

    `Args:`
        path: str
            Path of a temp file created by ``create_temp_file``
    `Returns:`
        bool
            Whether the temp file was found and closed
    """

    for temp_file in _temp_files:
        if temp_file.name == path:
            _temp_files.remove(temp_file)
            return True

    return False


def is_gzip_path(path):
    return (path[-3:] == '.gz')


def suffix_for_compression_type(compression):
    if compression == 'gzip':
        return '.gz'

    return ''


def compression_type_for_path(path):
    if is_gzip_path(path):
        return 'gzip'

    return None


def read_file(path):
    """
    Return the contents of file. Currently support `.gz` compressed files.

    `Args:`
        path: str
            The path to the file to read.
    `Returns:`
        str
            The contents of a files.
    """
    compression = compression_type_for_path(path)

    open_func = {
        'gzip': gzip.open,
        None: open,
    }
    with open_func[compression](path, 'r') as fp:
        return fp.read()


def string_to_temp_file(string, suffix=None):
    """
    Create a temporary file from a string. Currently used for packages
    that require credentials to be stored as a file.
    """

    temp_file_path = create_temp_file(suffix=suffix)

    with open(temp_file_path, 'w') as f:
        f.write(string)

    return temp_file_path


def zip_check(file_path, compression_type):
    """
    Check if the file suffix or the compression type indicates that it is
    a zip file.
    """

    if file_path:
        if file_path.split('/')[-1].split('.')[-1] == 'zip':
            return True

    if compression_type == 'zip':
        return True

    else:
        return False


def extract_file_name(file_path=None):
    """
    Extract the file name without the suffix from a file path string.
    """

    if not file_path:
        return None

    return file_path.split('/')[-1].split('.')[0]



import logging

logger = logging.getLogger(__name__)


class CanvassResponses(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_canvass_responses_contact_types(self):
        """
        Get canvass response contact types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'canvassResponses/contactTypes'

        crc = self.connection.request(url)
        logger.debug(crc)
        logger.info(f'Found {crc.num_rows} canvass response contact types.')

        return crc

    def get_canvass_responses_input_types(self):
        """
        Get canvass response input types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'canvassResponses/inputTypes'

        cri = self.connection.request(url)
        logger.debug(cri)
        logger.info(f'Found {cri.num_rows} canvass response input types.')

        return cri

    def get_canvass_responses_result_codes(self):
        """
        Get canvass response result code ids.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'canvassResponses/resultCodes'

        crr = self.connection.request(url)
        logger.debug(crr)
        logger.info(f'Found {crr.num_rows} canvass response result codes.')

        return crr





class SupporterGroups(object):
    """Class for '/activistCodes' end points."""

    def __init__(self, van_connection):
        """Initialize class"""

        self.connection = van_connection

    def get_supporter_groups(self):
        """
        Get supporter group objects

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'supporterGroups'

        return self.connection.request_paginate(url)

    def get_supporter_group(self, supporter_group_id):
        """
        Get a single supporter group object

        `Args:`
            supporter_group_id : int
                The activist code id associated with the activist code.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + \
            'supporterGroups/{}'.format(supporter_group_id)

        return self.connection.request(url)

    def create_supporter_group(self, name, description):
        """
        Create a new supporter group

        `Args:`
            name: str
                The name of the supporter group. 100 character limit
            description: str
                Optional; A description of the supporter group. 200 character limit.
        `Returns`
            Parsons Table with the newly createed supporter group id, name
            and description
        """

        url = self.connection.uri + 'supporterGroups'

        json = {'name': name, 'description': description}

        return self.connection.request(url, req_type="POST", post_data=json)

    def add_person_supporter_group(self, supporter_group_id, vanid):
        """
        Add a person to a supporter group

        `Args:`
            supporter_group_id: int
                The supporter group id
            vanid: int
                The vanid of the person to apply
        """

        url = self.connection.uri + \
            'supporterGroups/{}/people/{}'.format(supporter_group_id, vanid)

        return self.connection.request(url, req_type="PUT")

    def delete_person_supporter_group(self, supporter_group_id, vanid):
        """
        Remove a person from a supporter group

        `Args:`
            supporter_group_id: int
                The supporter group id
            vanid: int
                The vanid of the person to remove
        """

        url = self.connection.uri + \
            'supporterGroups/{}/people/{}'.format(supporter_group_id, vanid)

        return self.connection.request(url, req_type="DELETE")



"""NGPVAN Code Endpoints"""
import petl
from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class Codes(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_codes(self, name=None, supported_entities=None, parent_code_id=None,
                  code_type=None):
        """
        Get codes.

        `Args:`
            name : str
                Filter by name of code.
            supported_entities : str
                Filter by supported entities.
            parent_code_id : str
                Filter by parent code id.
            code_type : str
                Filter by code type.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        args = {'name': name,
                'supportedEntities': supported_entities,
                'parentCodeId': parent_code_id,
                'codeType': code_type,
                '$top': 200
                }

        url = self.connection.uri + 'codes'

        c = self.connection.request_paginate(url, args=args)
        logger.debug(c)
        logger.info(f'Found {c.num_rows} codes.')

        return c

    def get_code(self, code_id):
        """
        Get a code.

        `Args:`
            code_id : int
                The code id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'codes/{}'.format(code_id)

        c = self.connection.request(url)
        logger.debug(c)
        logger.info(f'Found code {code_id}.')

        return c

    def get_code_types(self):
        """
        Get code types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'codeTypes'

        ct = Table(petl.fromcolumns([self.connection.request(url, raw=True).json()],
                                    header=['code_type']))
        logger.debug(ct)
        logger.info(f'Found {ct.num_rows} code types.')

        return ct

    def create_code(self, name=None, parent_code_id=None, description=None,
                    code_type='SourceCode', supported_entities=None):
        """
        Create a code.

        `Args:`
            name: str
                The name of the code.
            parent_code_id: int
                A unique identifier for this code’s parent.
            description: str
                A description for this code, no longer than 200 characters.
            code_type: str
                The code type. ``Tag`` and ``SourceCode`` are valid values.
            supported_entities: list
                A list of dicts that enumerate the searchability and applicability rules of the
                code. You can find supported entities with the :meth:`code_supported_entities`

                .. highlight:: python
                .. code-block:: python

                    [
                        {
                         'name': 'Event',
                         'is_searchable': True,
                         'is_applicable': True
                        }
                        {
                         'name': 'Locations',
                         'start_time': '12-31-2018T13:00:00',
                         'end_time': '12-31-2018T14:00:00'
                        }
                    ]
        """

        url = self.connection.uri + 'codes'

        if supported_entities:

            se = [{'name': s['name'],
                   'isSearchable': s['is_searchable'],
                   'is_applicable': s['is_applicable']} for s in supported_entities]

        post_data = {"parentCodeId": parent_code_id,
                     "name": name,
                     "codeType": code_type,
                     "supportedEntities": se,
                     "description": description}

        c = self.connection.request(url, req_type="POST", post_data=post_data)
        logger.debug(c)
        logger.info(f'Code {c} created')
        return c

    def update_code(self, code_id, name=None, parent_code_id=None, description=None,
                    code_type='SourceCode', supported_entities=None):
        """
        Update a code.

        `Args:`
            code_id: int
                The code id.
            name: str
                The name of the code.
            parent_code_id: int
                A unique identifier for this code’s parent.
            description: str
                A description for this code, no longer than 200 characters.
            code_type: str
                The code type. ``Tag`` and ``SourceCode`` are valid values.
            supported_entities: list
                A list of dicts that enumerate the searchability and applicability rules of the
                code. You can find supported entities with the :meth:`code_supported_entities`

                .. highlight:: python
                .. code-block:: python

                    [
                        {
                         'name': 'Event',
                         'is_searchable': True,
                         'is_applicable': True
                        }
                        {
                         'name': 'Locations',
                         'start_time': '12-31-2018T13:00:00',
                         'end_time': '12-31-2018T14:00:00'
                        }
                    ]
        """

        url = self.connection.uri + 'codes/{}'.format(code_id)

        post_data = {}

        if name:
            post_data['name'] = name
        if parent_code_id:
            post_data['parentCodeId'] = parent_code_id
        if code_type:
            post_data['codeType'] = code_type
        if description:
            post_data['description'] = description

        if supported_entities:

            se = [{'name': s['name'],
                   'isSearchable': s['is_searchable'],
                   'is_applicable': s['is_applicable']} for s in supported_entities]
            post_data['supportedEntities'] = se

        c = self.connection.request(url, req_type="PUT", post_data=post_data)
        logger.debug(c)
        logger.info(f'Code {code_id} updated.')
        return c

    def delete_code(self, code_id):
        """
        Delete a code.

        `Args:`
            code_id: int
                The code id.
        `Returns:`
            ``(204, 'No Content')`` if successful
        """

        url = self.connection.uri + 'codes/{}'.format(code_id)

        c = self.connection.request(url, req_type="DELETE", raw=True)
        logger.info(f'Code {code_id} deleted.')
        return c

    def get_code_supported_entities(self):
        """
        Get code supported entities.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'codes/supportedEntities'

        cse = Table(petl.fromcolumns([self.connection.request(url, raw=True).json()],
                                     header=['supported_entities']))
        logger.debug(cse)
        logger.info(f'Found {cse.num_rows} code supported entities.')

        return cse





class Scores(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_scores(self):
        """
        Get all scores

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'scores'

        return self.connection.request_paginate(url)

    def get_score(self, score_id):
        """
        Get an individual score

        `Args:`
            score_id: int
                The score id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'scores/{}'.format(score_id)

        return self.connection.request(url)


class ScoreUpdates(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_score_updates(self, created_before=None, created_after=None,
                          score_id=None):
        """
        Get all score updates

        `Args:`
            created_before: str
                Optional; Filter score updates to those created before date. Use "YYYY-MM-DD"
                format.
            created_after: str
                Optional; Filter score updates to those created after date. Use "YYYY-MM-DD"
                format.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'scoreUpdates'

        args = {'createdBefore': created_before,
                'createdAfter': created_after,
                'scoreId': score_id}

        return self.connection.request_paginate(url, args=args)

    def get_score_update(self, score_update_id):
        """
        Get a score update object

            `Args:`
                score_update_id : int
                        The score update id
            `Returns:`
                Parsons Table
                    See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'scoreUpdates/{}'.format(score_update_id)

        return self.connection.request(url)

    def update_score_status(self, score_update_id, status):
        """
        Change the status of a score update object. This end point is used to
        approve a score loading job.

        `Args:`
            score_update_id: str
                The score update id
            status: str
                One of 'pending approval', 'approved', 'disapproved'
        `Returns:`
            boolean
                ``True`` if the status update is accepted
        """

        if status not in ['pending approval', 'approved', 'disapproved',
                          'canceled']:

            raise ValueError("""Valid inputs for status are, 'pending approval',
                             'approved','disapproved','canceled'""")

        else:
            if status == 'pending approval':
                status = 'PendingApproval'
            else:
                status = status.capitalize()

        post_data = {"loadStatus": status}

        url = self.connection.uri + 'scoreUpdates/{}'.format(score_update_id)

        r = self.connection.request(
            url, req_type="PATCH", post_data=post_data, raw=True)

        if r.status_code == 204:
            return True
        else:
            return r.status_code


class FileLoadingJobs(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def create_file_load(self, file_name, file_url, columns, id_column, id_type,
                         score_id, score_column, delimiter='csv', header=True,
                         description=None, email=None, auto_average=None,
                         auto_tolerance=None):
        """
        Loads a file. Only used for loading scores at this time. Scores must be
        compressed using `zip`.

        `Args:`
            file_name: str
                The name of the file contained in the zip file.
            file_url: str
                The url path to directly download the file. Can also be a path to an FTP site.
            columns: list
                A list of column names contained in the file.
            id_column: str
                The column name of the id column in the file.
            id_type: str
                A valid primary key, such as `VANID` or `DWID`. Varies by VAN instance.
            score_id: int
                The score slot id
            score_column: str
                    The column holding the score
            delimiter: str
                    The file delimiter used.
            email: str
                A valid email address in which file loading status will be sent.
            auto_average: float
                The average of scores to be loaded.
            auto_tolerance: float
                The fault tolerance of the VAN calculated average compared to the ``auto_average``.
                The tolerance must be less than 10% of the difference between the maximum and
                minimum possible acceptable values of the score.
        `Returns:`
            dict
                The file load id
        """

        columns = [{'name': c} for c in columns]

        # To Do: Validate that it is a .zip file. Not entirely sure if this is possible
        # as some urls might not end in ".zip".

        if delimiter not in ['csv', 'tab', 'pipe']:
            raise ValueError("Delimiter must be one of 'csv', 'tab' or 'pipe'")

        delimiter = delimiter.capitalize()

        url = self.connection.uri + 'FileLoadingJobs'

        post_data = {"description": 'A description',
                     "file": {
                         "columnDelimiter": delimiter,
                         "columns": columns,
                         "fileName": file_name,
                         "hasHeader": header,
                         "sourceUrl": file_url
                     },
                     "actions": [
                         {"actionType": "score",
                          "personIdColumn": id_column,
                          "personIdType": id_type,
                          "scoreColumn": score_column,
                          "scoreId": score_id}],
                     "listeners": [
                         {"type": "EMAIL",
                          "value": email}]
                     }

        if auto_average and auto_tolerance:

            post_data["actions"]["approvalCriteria"] = {"average": auto_average,
                                                        "tolerance": auto_tolerance}

        return self.connection.request(url, req_type="POST", post_data=post_data, raw=True)['jobId']

    def create_file_load_multi(self, file_name, file_url, columns, id_column, id_type,
                               score_map, delimiter='csv', header=True, description=None,
                               email=None):
        """
        An iteration of the :meth:`file_load` method that allows you to load multiple scores
        at the same time.

        `Args:`
            file_name : str
                The name of the file contained in the zip file.
            file_url : str
                The url path to directly download the file. Can also be a path to an FTP site.
            columns: list
                A list of column names contained in the file.
            id_column : str
                The column name of the id column in the file.
            id_type : str
                A valid primary key, such as `VANID` or `DWID`. Varies by VAN instance.
            score_map : list
                A list of dicts that adheres to the following syntax

                .. highlight:: python
                .. code-block:: python

                    [{'score_id' : int,
                      'score_column': str,
                      'auto_average': float,
                      'auto_tolerance': float }]

            email: str
                A valid email address in which file loading status will be sent.
        `Returns:`
            The file load job id
        """

        columns = [{'name': c} for c in columns]

        # To Do: Validate that it is a .zip file. Not entirely sure if this is possible
        # as some urls might not end in ".zip".

        if delimiter not in ['csv', 'tab', 'pipe']:
            raise ValueError("Delimiter must be one of 'csv', 'tab' or 'pipe'")

        delimiter = delimiter.capitalize()

        url = self.connection.uri + 'FileLoadingJobs'

        post_data = {"description": 'A description',
                     "file": {
                         "columnDelimiter": delimiter,
                         "columns": columns,
                         "fileName": file_name,
                         "hasHeader": header,
                         "sourceUrl": file_url
                     },

                     "listeners": [
                         {"type": "EMAIL",
                          "value": email}]
                     }

        actions = []

        for score in score_map:

            action = {"actionType": "score",
                      "personIdColumn": id_column,
                                    "personIdType": id_type,
                                    "scoreColumn": score['score_column'],
                                    "scoreId": score['score_id'],
                                    "approvalCriteria": {
                                        "average": score['auto_average'],
                                        "tolerance": score['auto_tolerance']
                                    }
                      }

            actions.append(action)

        post_data['actions'] = actions

        return self.connection.request(url, req_type="POST", post_data=post_data)



import unittest
import os
import requests_mock
from parsons.ngpvan.van import VAN
from test.utils import validate_list


os.environ['VAN_API_KEY'] = 'SOME_KEY'


class TestNGPVAN(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_connection(self, m):

        # Create response
        json = {u'message': u'True',
                u'dateSent': u'2018-08-11T15:08:20.1679461Z'}
        m.post(self.van.connection.uri + 'echoes/', json=json)

        # Tests API connection with a key.
        self.assertTrue(self.van.connection.api_test())

    @requests_mock.Mocker()
    def test_get_activist_codes(self, m):

        # Create response
        json = {u'count': 43, u'items':
                [{u'status': u'Active',
                  u'scriptQuestion': None,
                  u'name': u'TEST CODE',
                  u'mediumName': u'TEST CODE',
                  u'activistCodeId': 4388538,
                  u'shortName': u'TC',
                  u'type': u'Action',
                  u'description': None}],
                u'nextPageLink': None}

        m.get(self.van.connection.uri + 'activistCodes', json=json)

        # Expected Structure
        expected = ['status', 'scriptQuestion', 'name', 'mediumName',
                    'activistCodeId', 'shortName', 'type', 'description']

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.van.get_activist_codes()))

        # To Do: Test what happens when it doesn't find any ACs

    @requests_mock.Mocker()
    def test_get_activist_code(self, m):

        # Create response
        json = {"status": "Active",
                "scriptQuestion": "null",
                "name": "Anti-Choice",
                "mediumName": "Anti",
                "activistCodeId": 4135099,
                "shortName": "AC",
                "type": "Constituency",
                "description": "A person who has been flagged as anti-choice."}

        m.get(self.van.connection.uri + 'activistCodes/4388538', json=json)

        # Expected Structure
        expected = ['status', 'scriptQuestion', 'name', 'mediumName',
                    'activistCodeId', 'shortName', 'type', 'description']

        # Assert response is expected structure
        self.assertTrue(
            validate_list(
                expected,
                self.van.get_activist_code(4388538)))

        # To Do: Test what happens when it doesn't find the AC

    @requests_mock.Mocker()
    def test_get_canvass_responses_contact_types(self, m):

        json = {"name": "Auto Dial",
                "contactTypeId": 19,
                "channelTypeName": "Phone"}

        m.get(
            self.van.connection.uri +
            'canvassResponses/contactTypes',
            json=json)

        # Expected Structure
        expected = ['name', 'contactTypeId', 'channelTypeName']

        self.assertTrue(
            validate_list(
                expected,
                self.van.get_canvass_responses_contact_types()))

    @requests_mock.Mocker()
    def test_get_canvass_responses_input_types(self, m):

        json = {"inputTypeId": 11, "name": "API"}

        m.get(
            self.van.connection.uri +
            'canvassResponses/inputTypes',
            json=json)

        # Expected Structure
        expected = ['inputTypeId', 'name']

        self.assertTrue(
            validate_list(
                expected,
                self.van.get_canvass_responses_input_types()))

    @requests_mock.Mocker()
    def test_get_canvass_responses_result_codes(self, m):

        json = {
            "shortName": "BZ",
            "resultCodeId": 18,
            "name": "Busy",
            "mediumName": "Busy"
        }

        m.get(
            self.van.connection.uri +
            'canvassResponses/resultCodes',
            json=json)

        # Expected Structure
        expected = ['shortName', 'resultCodeId', 'name', 'mediumName']

        self.assertTrue(
            validate_list(
                expected,
                self.van.get_canvass_responses_result_codes()))

    @requests_mock.Mocker()
    def test_get_saved_lists(self, m):

        json = {'count': 1, 'items': [
            {"savedListId": 517612,
             "listCount": 974656,
             "name": "LikelyParents(16andunder)_DWID_S... - MN",
             "doorCount": 520709,
             "description": "null"
             }
        ], 'nextPageLink': None}

        m.get(self.van.connection.uri + 'savedLists', json=json)

        expected = [
            'savedListId',
            'listCount',
            'name',
            'doorCount',
            'description']

        self.assertTrue(validate_list(expected, self.van.get_saved_lists()))

    @requests_mock.Mocker()
    def test_get_saved_list(self, m):

        saved_list_id = 517612

        json = {"savedListId": 517612,
                "listCount": 974656,
                "name": "LikelyParents(16andunder)_DWID_S... - MN",
                "doorCount": 520709,
                "description": "null"
                }

        m.get(
            self.van.connection.uri +
            'savedLists/{}'.format(saved_list_id),
            json=json)

        expected = [
            'savedListId',
            'listCount',
            'name',
            'doorCount',
            'description']

        self.assertTrue(
            validate_list(
                expected,
                self.van.get_saved_list(saved_list_id)))

    @requests_mock.Mocker()
    def test_folders(self, m):

        json = {u'count': 2,
                u'items': [
                    {
                        u'folderId': 5046,
                        u'name': u'#2018_MN_active_universe'
                    },
                    {u'folderId': 2168,
                     u'name': u'API Generated Lists'
                     }
                ], u'nextPageLink': None}

        m.get(self.van.connection.uri + 'folders', json=json)

        expected = ['folderId', 'name']

        self.assertTrue(validate_list(expected, self.van.folders()))

    @requests_mock.Mocker()
    def test_folder(self, m):

        folder_id = 5046

        json = {"folderId": 5046, "name": "#2018_MN_active_universe"}

        m.get(
            self.van.connection.uri +
            'folders/{}'.format(folder_id),
            json=json)

        expected = ['folderId', 'name']

        self.assertTrue(validate_list(expected, self.van.folder(folder_id)))

    @requests_mock.Mocker()
    def test_export_job_types(self, m):

        json = {u'count': 1, u'items':
                [{u'exportJobTypeId': 4, u'name': u'SavedListExport'}],
                u'nextPageLink': None}

        m.get(self.van.connection.uri + 'exportJobTypes', json=json)

        expected = ['exportJobTypeId', 'name']

        self.assertTrue(validate_list(expected, self.van.export_job_types()))

    @requests_mock.Mocker()
    def test_export_job_create(self, m):

        saved_list_id = 517612

        json = {"status": "Completed",
                "errorCode": "null",
                "exportJobGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "activistCodes": "null",
                "canvassFileRequestId": 448,
                "dateExpired": "2018-09-08T16:04:00Z",
                "surveyQuestions": "null",
                "webhookUrl": "https://www.nothing.com/",
                "downloadUrl": "https://ngpvan.blob.core.windows.net/canvass-files-savedlistexport/bf4d1297-1c77-3fb2-03bd-f0acda122d37_2018-09-08T13:03:27.7191831-04:00.csv",  # noqa: E501
                "savedListId": 517612,
                "districtFields": "null",
                "canvassFileRequestGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "customFields": "null",
                "type": 4,
                "exportJobId": 448}

        m.post(self.van.connection.uri + 'exportJobs', json=json)

        expected = [
            'status',
            'errorCode',
            'exportJobGuid',
            'activistCodes',
            'canvassFileRequestId',
            'dateExpired',
            'surveyQuestions',
            'webhookUrl',
            'downloadUrl',
            'savedListId',
            'districtFields',
            'canvassFileRequestGuid',
            'customFields',
            'type',
            'exportJobId']

        self.assertEqual(json,self.van.export_job_create(saved_list_id))

    @requests_mock.Mocker()
    def test_export_job(self, m):

        export_job_id = 448

        json = {"status": "Completed",
                "errorCode": "null",
                "exportJobGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "activistCodes": "null",
                "canvassFileRequestId": 448,
                "dateExpired": "2018-09-08T16:04:00Z",
                "surveyQuestions": "null",
                "webhookUrl": "https://www.nothing.com/",
                "downloadUrl": "https://ngpvan.blob.core.windows.net/canvass-files-savedlistexport/bf4d1297-1c77-3fb2-03bd-f0acda122d37_2018-09-08T13:03:27.7191831-04:00.csv",  # noqa: E501
                "savedListId": 517612,
                "districtFields": "null",
                "canvassFileRequestGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "customFields": "null",
                "type": 4,
                "exportJobId": 448}

        expected = [
            'status',
            'errorCode',
            'exportJobGuid',
            'activistCodes',
            'canvassFileRequestId',
            'dateExpired',
            'surveyQuestions',
            'webhookUrl',
            'downloadUrl',
            'savedListId',
            'districtFields',
            'canvassFileRequestGuid',
            'customFields',
            'type',
            'exportJobId']

        m.get(
            self.van.connection.uri +
            'exportJobs/{}'.format(export_job_id),
            json=json)

        self.assertTrue(validate_list(expected, self.van.export_job(export_job_id)))

    @requests_mock.Mocker()
    def test_get_survey_questions(self, m):

        json = {u'count': 67, u'items': [{
            "status": "Active",
            "responses": [
                {"shortName": "1",
                 "surveyResponseId": 1288926,
                 "name": "1-Strong Walz",
                         "mediumName": "1"},
                {"shortName": "2",
                 "surveyResponseId": 1288928,
                 "name": "2-Lean Walz",
                         "mediumName": "2"}],
            "scriptQuestion": "Who do you support for Governor?",
            "name": "MN Governor Gen",
                    "surveyQuestionId": 311838,
                    "mediumName": "MNGovG",
                    "shortName": "MGG",
                    "type": "Candidate",
                    "cycle": 2018
        }],
            u'nextPageLink': None}

        m.get(self.van.connection.uri + 'surveyQuestions', json=json)

        expected = ['status', 'responses', 'scriptQuestion', 'name',
                    'surveyQuestionId', 'mediumName', 'shortName',
                    'type', 'cycle']

        self.assertTrue(validate_list(expected, self.van.get_survey_questions()))

    @requests_mock.Mocker()
    def test_get_supporter_groups(self, m):

        json = {"items": [
            {
                "id": 12,
                "name": "tmc",
                "description": "A fun group."
            },
            {
                "id": 13,
                "name": "tmc",
                "description": "A fun group."
            },
        ],
            "nextPageLink": None,
            "count": 3
        }

        m.get(self.van.connection.uri + 'supporterGroups', json=json)

        expected = ['id', 'name', 'description']

        self.van.get_supporter_groups()

    @requests_mock.Mocker()
    def test_get_supporter_group(self, m):

        supporter_group_id = 12

        json = {"id": 12, "name": "tmc", "description": "A fun group."}

        m.get(self.van.connection.uri + 'supporterGroups/{}'.format(supporter_group_id), json=json)

        expected = ['id', "name", "description"]

        r = self.van.get_supporter_group(supporter_group_id)

        # Test that columns are expected columns
        self.assertTrue(validate_list(expected, r))

        # Test correct group id
        self.assertEqual(supporter_group_id, r[0]['id'])

    @requests_mock.Mocker()
    def test_add_person_supporter_group(self, m):

        good_supporter_group_id = 5
        good_vanid = 12345

        bad_supporter_group_id = 999
        bad_vanid = 99999

        good_ep = 'supporterGroups/{}/people/{}'.format(good_vanid, good_supporter_group_id)

        m.put(self.van.connection.uri + good_ep, status_code=204)

        bad_ep = 'supporterGroups/{}/people/{}'.format(bad_vanid, bad_supporter_group_id)

        m.put(self.van.connection.uri + bad_ep, status_code=404)

        # Test bad input
        self.assertEqual(self.van.add_person_supporter_group(bad_vanid, bad_supporter_group_id)[0], 404)

        # Test good input
        self.assertEqual(self.van.add_person_supporter_group(good_vanid, good_supporter_group_id)[0], 204)

    @requests_mock.Mocker()
    def test_delete_person_supporter_group(self, m):

        good_supporter_group_id = 5
        good_vanid = 12345

        bad_supporter_group_id = 999
        bad_vanid = 99999

        good_ep = 'supporterGroups/{}/people/{}'.format(good_vanid, good_supporter_group_id)

        m.delete(self.van.connection.uri + good_ep, status_code=204)

        bad_ep = 'supporterGroups/{}/people/{}'.format(bad_vanid, bad_supporter_group_id)

        m.delete(self.van.connection.uri + bad_ep, status_code=404)

        # Test bad input
        self.assertEqual(self.van.delete_person_supporter_group(bad_vanid, bad_supporter_group_id)[0], 404)

        # Test good input
        self.assertEqual(self.van.delete_person_supporter_group(good_vanid, good_supporter_group_id)[0], 204)

    @requests_mock.Mocker()
    def test_get_scores(self, m):

        json = {u'count': 2, u'items':
                [{u'origin': None,
                  u'scoreId': 2716,
                  u'name': u'Democratic Party Support',
                  u'maxValue': 100.0,
                  u'minValue': 1.0,
                  u'state': None,
                  u'shortName': u'Dem Support',
                  u'description': None}],
                u'nextPageLink': None}

        m.get(self.van.connection.uri + 'scores', json=json)

        expected = ['origin', 'scoreId', 'name', 'maxValue', 'minValue',
                    'state', 'shortName', 'description']

        self.assertTrue(validate_list(expected, self.van.get_scores()))

    @requests_mock.Mocker()
    def test_get_score(self, m):

        score_id = 2716

        json = {u'origin': None,
                u'scoreId': 2716,
                u'name': u'Democratic Party Support',
                u'maxValue': 100.0,
                u'minValue': 1.0,
                u'state': None,
                u'shortName': u'Dem Support',
                u'description': None}

        m.get(self.van.connection.uri + 'scores/{}'.format(score_id), json=json)

        expected = ['origin', 'scoreId', 'name', 'maxValue', 'minValue',
                    'state', 'shortName', 'description']

        self.assertTrue(validate_list(expected, self.van.get_score(score_id)))

    @requests_mock.Mocker()
    def test_get_score_updates(self, m):

        json = {'count': 46, 'items':
                [{
                    "loadStatus": "Canceled",
                    "updateStatistics": {
                        "increasedBy": 1,
                        "nulledOut": 1,
                        "added": 0,
                        "matchedRows": 4,
                        "matchPercent": 100.0,
                        "outOfRange": 0,
                        "badValues": 1,
                        "totalRows": 4,
                        "maxValue": 30.0,
                        "medianValue": 15.0,
                        "minValue": 10.0,
                        "duplicateRows": "null",
                        "averageValue": 20.0,
                        "decreasedBy": 2
                    },
                    "score": {
                        "origin": "null",
                        "scoreId": 2716,
                        "name": "Democratic Party Support",
                        "maxValue": 100.0,
                        "minValue": 1.0,
                        "state": "null",
                        "shortName": "null",
                        "description": "null"
                    },
                    "dateProcessed": "null",
                    "scoreUpdateId": 27892
                }],
                'nextPageLink': None}

        m.get(self.van.connection.uri + 'scoreUpdates', json=json)

        expected = ['loadStatus', 'updateStatistics', 'score', 'dateProcessed', 'scoreUpdateId']

        self.assertTrue(validate_list(expected, self.van.get_score_updates()))

    @requests_mock.Mocker()
    def test_get_score_update(self, m):

        score_update_id = 27892

        json = {
            "loadStatus": "Canceled",
            "updateStatistics": {
                "increasedBy": 1,
                "nulledOut": 1,
                "added": 0,
                "matchedRows": 4,
                "matchPercent": 100.0,
                "outOfRange": 0,
                "badValues": 1,
                "totalRows": 4,
                "maxValue": 30.0,
                "medianValue": 15.0,
                "minValue": 10.0,
                "duplicateRows": "null",
                "averageValue": 20.0,
                "decreasedBy": 2
            },
            "score": {
                "origin": "null",
                "scoreId": 2716,
                "name": "Democratic Party Support",
                        "maxValue": 100.0,
                        "minValue": 1.0,
                        "state": "null",
                        "shortName": "null",
                        "description": "null"
            },
            "dateProcessed": "null",
            "scoreUpdateId": 27892
        }

        m.get(self.van.connection.uri + 'scoreUpdates/{}'.format(score_update_id), json=json)

        expected = ['loadStatus', 'updateStatistics', 'score', 'dateProcessed', 'scoreUpdateId']

        self.assertTrue(validate_list(expected, self.van.get_score_update(score_update_id)))

    @requests_mock.Mocker()
    def test_update_score_status(self, m):

        score_update_id = 27892

        m.patch(self.van.connection.uri + 'scoreUpdates/{}'.format(score_update_id),
                status_code=204)

        # Test bad input
        self.assertRaises(ValueError, self.van.update_score_status, score_update_id, 'not a thing.')

        # Test good input
        self.assertTrue(self.van.update_score_status(score_update_id, 'approved'))

    @requests_mock.Mocker()
    def test_create_file_load(self, m):

        file_name = 'test_scores.csv'
        file_url_good = 'http://tmc.org/test_scores.zip'
        file_url_bad = 'http://tmc.org/test_scores'
        columns = ['vanid', 'score']
        id_column = 'vanid'
        id_type = 'VANID'
        score_id = 2716
        score_column = 'score'
        bad_delimiter = '*'

        json = {'jobId': 9749}

        m.post(self.van.connection.uri + 'FileLoadingJobs', json=json)

        # Test bad delimiter
        self.assertRaises(ValueError, self.van.create_file_load, file_name, file_url_good, columns,
                          id_column, id_type, score_id, score_column, delimiter=bad_delimiter)

        # Test good request
        self.assertEqual(json['jobId'], self.van.create_file_load(file_name, file_url_good, columns,
                                                                  id_column, id_type, score_id,
                                                                  score_column))

    # TO DO: Test Volunteer Activites


if __name__ == '__main__':

    unittest.main()



class SurveyQuestions(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_survey_questions(self, statuses=None, name=None,
                             sq_type=None, question=None, cycle=None,
                             page_size=200):
        """
        Get survey questions.

        `Args:`
            statuses: str
                Comma delimited list of statuses of Survey Questions. One or
                more of Active (default), Archived, and Inactive.
            name: str
                Filter to survey questions with names begin with the input
            sq_type: str
                Filter to survey questions of the given type
            question: str
                Filter to survey questions with script questions that
                contain the given input
            cycle: str
                A year in the format YYYY; filters to Survey Questions with
                the given cycle
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'surveyQuestions'

        args = {'statuses': statuses,
                '$top': page_size,
                'name': name,
                'type': sq_type,
                'question': question,
                'cycle': cycle
                }

        return self.connection.request_paginate(url, args=args)

    def get_survey_question(self, survey_question_id):
        """
        Get a survey question

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + \
            'surveyQuestions/{}'.format(survey_question_id)

        return self.connection.request(url)



from requests import request as _request
from parsons.etl.table import Table
import logging
from parsons.utilities import check_env

logger = logging.getLogger(__name__)

URI = 'https://api.securevan.com/v4/'


class VANConnector(object):

    def __init__(self, api_key=None, auth_name='default', db=None, raise_for_status=True):

        self.api_key = check_env.check('VAN_API_KEY', api_key)

        if db == 'MyVoters':
            self.db_code = 0
        elif db in ['MyMembers', 'MyCampaign', 'EveryAction']:
            self.db_code = 1
        else:
            raise KeyError('Invalid database type specified. Pick one of:'
                           ' MyVoters, MyCampaign, MyMembers, EveryAction.')

        self.uri = URI
        self.db = db
        self.auth_name = auth_name
        self.raise_for_status = raise_for_status
        self.auth = (self.auth_name, self.api_key + '|' + str(self.db_code))

    def _error_check(self, r):

        if r.status_code == 404:

            # To Do: Make the errors prettier...
            logger.info(f"{r.status_code} Error: {r.json()['errors']}")

            return r.json()['errors']

        else:

            logger.debug(f'{r.json()}')
            return r.json()

    def request(self, url, req_type='GET', post_data=None, args=None, raw=False, paginate=False):
        # Internal request function

        r = _request(req_type, url, auth=self.auth, json=post_data, params=args)

        """
        # To Do: Figure out if this is still needed.
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as err:
            logging.info(err)
            if self.raise_for_status:
                r.raise_for_status()
        """

        if req_type in ['POST', 'DELETE', 'PUT'] or (req_type == 'GET' and r.status_code != 200):
            return self.code_parse(r)

        if raw:
            # Commenting out for the moment
            # return self._error_check(r)
            return r

        elif paginate:
            return r.json()

        else:
            if len(r.text) == 0:
                data = []
            elif isinstance(r.json(), list):
                data = r.json()
            else:
                data = [r.json()]

            if not data:
                logging.warning('No data returned in table.')

            return Table(data)

    def request_paginate(self, url, req_type='GET', post_data=None, args=None):
        # Internal request function that paginates

        items = []
        next_page = True

        while next_page:

            i = self.request(url, req_type=req_type, post_data=post_data, args=args, paginate=True)

            if 'count' not in i or i['count'] == 0:
                return Table(items)

            items.extend(i['items'])

            if not i['nextPageLink']:
                next_page = False

            url = i['nextPageLink']

        return Table(items)

    def api_test(self):

        url = self.uri + 'echoes/'
        r = self.request(url, req_type="POST", post_data={'message': 'True'})
        if r['message'] == 'True':
            return True
        else:
            return False

    def code_parse(self, req_obj):

        if req_obj.status_code == 200 and len(req_obj.text) == 0:
            return (200, 'OK')

        elif req_obj.status_code == 204:
            return (204, 'No Content')

        elif req_obj.status_code == 404:
            return (404, 'Not Found')

        elif req_obj.status_code in [400, 403, 500]:
            return (int(req_obj.status_code), req_obj.json())

        else:
            return req_obj.json()



import logging

logger = logging.getLogger(__name__)


class Locations(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_locations(self, name=None):
        """
        Get locations.

        `Args:`
            name: str
                Filter locations by name.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
         """

        url = self.connection.uri + 'locations'

        if name:
            name = {'name': name}

        return self._unpack_loc(self.connection.request_paginate(url, args=name))

    def get_location(self, location_id):
        """
        Get a location.

        `Args:`
            location_id: int
                The location id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'locations/{}'.format(location_id)

        return self._unpack_loc(self.connection.request(url))

    def create_location(self, name, address_line1=None, address_line2=None, city=None,
                        state=None, zip_code=None):
        """
        Find or create a location. If location already exists, will return location id.

        `Args:`
            name: str
                A name for this location, no longer than 50 characters.
            address_line1: str
                First line of a street address.
            address_line2: str
                Second line of a street address.
            city: str
                City or town name.
            state: str
                Two or three character state or province code (e.g., MN, ON, NSW, etc.).
            zip_code: str
                ZIP, ZIP+4, Postal Code, Post code, etc.
            `Returns:`
                A location id
        """

        url = self.connection.uri + 'locations/findOrCreate'

        location = {'name': name,
                    'address': {
                        'addressLine1': address_line1,
                        'addressLine2': address_line2,
                        'city': city,
                        'stateOrProvince': state,
                        'zipOrPostalCode': zip_code
                    }}

        return self.connection.request(url, req_type='POST', post_data=location, raw=True)

    def delete_location(self, location_id):
        """
        Delete a location.

        `Args:`
            location_id: int
                The location id
        `Returns:`
            ``200: OK`` if successful and ``404 Not Found`` if location not found
        """

        url = self.connection.uri + 'locations/{}'.format(location_id)

        return self.connection.request(url, req_type='DELETE', raw=True)

    def _unpack_loc(self, table):
        # Internal method to unpack location json

        if isinstance(table, tuple):
            return table

        if 'address' in table.columns:
            table.unpack_dict('address', prepend=False)

        if 'geoLocation' in table.columns:
            table.unpack_dict('geoLocation', prepend=False)

        return table



from parsons.etl.table import Table


class SavedLists(object):

    def __init__(self, van_connection):

        # Some sort of test if the van_connection is not present.

        self.connection = van_connection

    def get_saved_lists(self, folder_id=None):
        """
        Get all saved lists

        `Args:`
            folder_id : int
                Optional; the id for a VAN folder. If included returns only
                the saved lists in the folder
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'savedLists'

        return self.connection.request_paginate(
            url, args={'folderId': folder_id})

    def get_saved_list(self, saved_list_id):
        """
        Returns a single saved list object

        `Args:`
            saved_list_id : int
                The saved list id associated with the saved list.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'savedLists/{}'.format(str(saved_list_id))

        return self.connection.request(url)

    def download_saved_list(self, saved_list_id):
        """
        Download a saved list

        `Args:`
            saved_list_id
                The saved list id associated with the saved list.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        ej = ExportJobs(self.connection)
        job = ej.export_job_create(saved_list_id)

        if isinstance(job, tuple):
            return job
        else:
            return Table.from_csv(job['downloadUrl'])


class Folders(object):

    def __init__(self, van_connection):

        # Some sort of test if the van_connection is not present.

        self.connection = van_connection

    def folders(self):
        """
        List all folders

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'folders'

        return self.connection.request_paginate(url)

    def folder(self, folder_id):
        """
        Get a single folder

        `Args:`
            folder_id: int
                The folder id associated with the folder.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'folders/{}'.format(str(folder_id))

        return self.connection.request(url)


class ExportJobs(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def export_job_types(self):
        """
        Lists export job types

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'exportJobTypes'

        return self.connection.request_paginate(url)

    def export_job_create(self, list_id, export_type=4,
                          webhookUrl="https://www.nothing.com"):
        """
        Creates an export job

        Currently, this is only used for exporting saved lists. It is
        recommended that you use the :meth:`saved_list_download` method
        instead.

        `Args:`
            list_id: int
                This is where you should input the list id
            export_type: int
                The export type id, which defines the columns to export
            webhookUrl:
                A webhook to include to notify as to the status of the export
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options. Includes a
                download link for the file.
        """

        url = self.connection.uri + 'exportJobs'

        data = {"savedListId": str(list_id),
                "type": str(export_type),
                "webhookUrl": webhookUrl
                }

        return self.connection.request(url, req_type='POST', post_data=data, raw=True)

    def export_job(self, export_job_id):
        """
        Returns a single export job

        `Args:`
            export_job_id: int
                Export job id
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.connection.uri + 'exportJobs/{}'.format(str(export_job_id))

        return self.connection.request(url)
